/*
 * This file is part of FGGL.
 *
 * FGGL is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any
 * later version.
 *
 * FGGL is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with FGGL.
 * If not, see <https://www.gnu.org/licenses/>.
 */

#include <utility>

#include "fggl/display/glfw/window.hpp"
#include "fggl/display/glfw/window_input.hpp"
#include "fggl/debug/logging.hpp"

#include <string>
#include <stdexcept>
#include <GLFW/glfw3.h>

namespace fggl::display::glfw {

	static void glfw_error(int code, const char *description) {
		debug::log(debug::Level::error, "[GLFW] {}: {}", code, description);
	}

	static void framebuffer_resize(GLFWwindow *window, int width, int height) {
		auto *fgglWindow = static_cast<Window *>(glfwGetWindowUserPointer(window));
		fgglWindow->framesize(width, height);
	}

	static void fggl_input_cursor(GLFWwindow *window, double xPos, double yPos) {
		auto &input = GlfwInputManager::instance();
		auto *fgglWin = static_cast<Window *>(glfwGetWindowUserPointer(window));

		#ifndef FGGL_INPUT_SCREEN_COORDS
		// convert to nice ranges...
		xPos = (xPos / fgglWin->width() * 2) - 1.0; // [-1, 1]
		yPos = (yPos / fgglWin->height() * 2) - 1.0; // [-1, 1]
		#endif

		// inform the input system
		input.onMouseMove(xPos, yPos);
	}

	static void fggl_input_scroll(GLFWwindow */*window*/, double xPos, double yPos) {
		auto &input = GlfwInputManager::instance();
		input.onMouseScroll(xPos, yPos);
	}

	static void fggl_input_mouse_btn(GLFWwindow */*window*/, int btn, int action, int /*mods*/) {
		auto &input = GlfwInputManager::instance();
		input.onMouseButton(btn, action == GLFW_PRESS);
	}

	static void fggl_input_keyboard(GLFWwindow */*window*/, int /*key*/, int scancode, int action, int /*mods*/) {
		auto &input = GlfwInputManager::instance();
		input.onKeyEvent(scancode, action == GLFW_PRESS || action == GLFW_REPEAT);
	}

	static void fggl_update_joystick(fggl::input::GamepadInput &input, int jid) {
		bool isGamepad = (glfwJoystickIsGamepad(jid) == GLFW_TRUE);

		if (isGamepad) {
			if (!input.present(jid)) {
				std::string name = glfwGetJoystickName(jid);
				input.setActive(jid, true);
				input.name(jid, name);
			}

			GLFWgamepadstate state;
			glfwGetGamepadState(jid, &state);

			fggl::input::GamepadState gamepadState;
			for (int i = 0; i <= GLFW_GAMEPAD_BUTTON_LAST; i++) {
				gamepadState.buttons[i] = state.buttons[i];
			}

			for (int i = 0; i <= GLFW_GAMEPAD_AXIS_LAST; i++) {
				gamepadState.axes[i] = state.axes[i];
			}

			input.update(jid, gamepadState);
		} else {
			input.setActive(jid, false);
		}
	}

	static void fggl_joystick_poll() {
		auto &input = GlfwInputManager::instance();
		if (!input.alive()) {
			return;
		}

		auto &gamepadCtl = input.gamepads();
		for (int jid = 0; jid < GLFW_JOYSTICK_LAST; jid++) {
			if (glfwJoystickPresent(jid) == GLFW_TRUE) {
				fggl_update_joystick(gamepadCtl, jid);
			} else {
				gamepadCtl.setActive(jid, false);
			}
		}
	}

	static void fggl_joystick(int jid, int state) {
		auto &input = GlfwInputManager::instance();
		if (!input.alive()) {
			return;
		}

		auto &gamepadCtl = input.gamepads();
		if (state == GLFW_CONNECTED) {
			fggl_update_joystick(gamepadCtl, jid);
		} else {
			gamepadCtl.setActive(jid, false);
		}
	}

	GlfwContext::GlfwContext(fggl::input::Input *input) {
		debug::log(debug::Level::info, "[GLFW] Initializing...");
		auto &glfwCallbacks = GlfwInputManager::instance();
		glfwCallbacks.setup(input);

		glfwInitHint(GLFW_JOYSTICK_HAT_BUTTONS, GLFW_FALSE);
		glfwSetErrorCallback(glfw_error);

		int state = glfwInit();
		if (state == GLFW_FALSE) {
			const char **error = nullptr;
			glfwGetError(error);
			throw std::runtime_error(*error);
		}

		// joysticks are global
		fggl_joystick_poll();
		glfwSetJoystickCallback(fggl_joystick);
		debug::log(debug::Level::info, "[GLFW] Initialization complete.");
	}

	GlfwContext::~GlfwContext() {
		glfwTerminate();
		debug::trace("[glfw] context terminated");
	}

	void GlfwContext::pollEvents() {
		auto &glfwCallbacks = GlfwInputManager::instance();
		glfwCallbacks.frame();

		glfwPollEvents();
		fggl_joystick_poll();
	}

	Window::Window(std::shared_ptr<GlfwContext> context, gfx::WindowGraphics *graphics)
		: m_context(std::move(context)), m_window(nullptr), m_framesize() {

		// don't iconify when focus is lost.
		glfwWindowHint( GLFW_AUTO_ICONIFY, GLFW_FALSE );

		// FIXME - this ties the graphics API before window creation
		auto graphicsConfig = graphics->config();
		if (graphicsConfig.api == gfx::GraphicsAPI::OpenGL) {
			glfwWindowHint(GLFW_CONTEXT_VERSION_MAJOR, graphicsConfig.majorVersion);
			glfwWindowHint(GLFW_CONTEXT_VERSION_MINOR, graphicsConfig.minorVersion);
			glfwWindowHint(GLFW_OPENGL_PROFILE, GLFW_OPENGL_CORE_PROFILE);
			glfwWindowHint(GLFW_OPENGL_DEBUG_CONTEXT, graphicsConfig.debug ? GLFW_TRUE : GLFW_FALSE);
		}

		m_window = glfwCreateWindow(1920, 1080, "main", nullptr, nullptr);
		if (m_window == nullptr) {
			return;
		}

		glfwSetWindowUserPointer(m_window, this);

		// setup callback wrappers (will invoke the methods via the user ptr above)
		glfwSetFramebufferSizeCallback(m_window, framebuffer_resize);
		glfwSetScrollCallback(m_window, fggl_input_scroll);
		glfwSetCursorPosCallback(m_window, fggl_input_cursor);
		glfwSetMouseButtonCallback(m_window, fggl_input_mouse_btn);
		glfwSetKeyCallback(m_window, fggl_input_keyboard);

		// bind the graphics API
		glfwMakeContextCurrent(m_window);
		m_graphics = graphics->createMain(*this);
	}

	Window::~Window() {
		if ( m_graphics != nullptr ) {
			m_graphics = nullptr;
		}

		if (m_window != nullptr) {
			// prevent dangling pointers
			glfwSetWindowUserPointer(m_window, nullptr);
			glfwDestroyWindow(m_window);
			m_window = nullptr;
		}
	}

	void Window::frameStart() {
		assert(m_window != nullptr);
		assert(m_graphics != nullptr);

		glfwMakeContextCurrent(m_window);
		m_graphics->clear();
	}

	void Window::frameEnd() {
		glfwSwapBuffers(m_window);
	}

	void Window::activate() const {
		assert(m_window != nullptr);
		glfwMakeContextCurrent(m_window);
	}

	fggl::math::vec2i Window::frameSize() const {
		assert(m_window != nullptr);
		math::vec2i size;
		glfwGetFramebufferSize(m_window, &size.x, &size.y);
		return size;
	}

	bool Window::wantClose() const {
		assert(m_window != nullptr);
		return glfwWindowShouldClose(m_window) == GLFW_TRUE;
	}
}